/*
 * Meta, Geoffrey's self-balancing ride-on robot.
 * 
 * Copyright (C) 2006-2007 Geoffrey D. Bennett <g@netcraft.com.au>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <avr/eeprom.h>

#include "config.h"

const struct config_struct default_config = {
  /* ADXRS401 -- 15mV/deg/sec (5V, 10 bit ADC) */
  /* pitch_rate_conv */ 5.0 / 1024.0 * 1000 / 15 / 57.29578,

  /* ADXL203 -- 5g across full scale (5V, 10bit ADC) */
  /* ay_conv */ 5.0 / 1024.0,

  /* ay_bal_offset */ 0.3,

  /* 50V across full scale, (10 bit resolution, 5V, *10 divider)  */
  /* batt_conv */ /*1.0 / 1024.0 * 5.0 * 10.0, */ 0.05,

  /* p_gain */ 2.0,
  /* d_gain */ 0.1,
  /* i_gain */ 3.0,

  /* hard_speed_lim */ 0.7,
  /* speed_limit_fwd */ 0.7,
  /* speed_limit_rev */ 0.7,
  /* overspeed_p_gain */ 0.5,
  /* overspeed_i_gain */ 1.0,
  /* not_overspeed_i_gain */ 2.0,
  /* overspeed_angle_limit */ 0.2,

  /* softstart_delay */ 0.3,
  /* tip_angle */ 0.35,

  /* turn_range */ 410,
  /* turn_gain */ 0.1,
  /* turn_backoff */ 10,
  /* turn_filter */ 0.05,

  /* rc_throttle_midpoint */ 94,
  /* rc_throttle_range */ 50,
  /* rc_throttle_angle_conv */ 0.1,
  /* rc_throttle_filter */ 0.05,

  /* rc_steer_midpoint */ 94,
  /* rc_steer_range */ 50,
  /* rc_steer_conv */ 0.1,

  /* motor_l_gain */ 0,
  /* motor_r_gain */ 0,
};

union config_union config;

uint8_t EEMEM eeprom_version;
uint8_t EEMEM eeprom_length;
uint16_t EEMEM eeprom_checksum;
struct config_struct EEMEM eeprom_config;

static uint16_t calculate_checksum(uint16_t *data) {
  uint16_t sum = 0;
  int i;

  for (i = 0; i < CONFIG_WORD_LEN; i++) {
    sum += *data++;
  }

  return sum;
}

void config_load_defaults(void) {
  config.s = default_config;
}

void eeprom_save_config(void) {
  eeprom_write_byte(&eeprom_length, 0);
  eeprom_write_byte(&eeprom_version, EEPROM_VERSION);
  eeprom_write_word(&eeprom_checksum, calculate_checksum(config.c));
  eeprom_write_block((const void *) &config.s,
                     (void *) &eeprom_config,
                     sizeof(struct config_struct));
  eeprom_write_byte(&eeprom_length, NUM_CONFIG_VAR);
}

/* returns true for config loaded succesfully from EEPROM, false if
 * EEPROM was invalid and defaults were restored
 */
int eeprom_load_config(void) {
  int valid_config = 1;

  log_str("Check EEPROM\r\n");

  if (eeprom_read_byte(&eeprom_version) != EEPROM_VERSION) {
    log_str("Bad version\r\n");
    valid_config = 0;
  } else if (eeprom_read_byte(&eeprom_length) != NUM_CONFIG_VAR) {
    log_str("Bad config length\r\n");
    valid_config = 0;
  } else {
    uint16_t stored_checksum = eeprom_read_word(&eeprom_checksum);
    eeprom_read_block((void *) &config.s,
                      (const void *) &eeprom_config,
                      sizeof(struct config_struct));
    if (stored_checksum != calculate_checksum(config.c)) {
      log_str("Bad checksum\r\n");
      valid_config = 0;
    } else {
      log_str("Config loaded\r\n");
    }
  }

  if (!valid_config) {
    log_str("Setting defaults\r\n");
    uart_flush();
    config_load_defaults();
    eeprom_save_config();
  }

  return valid_config;
}
