/*
 * Meta, Geoffrey's self-balancing ride-on robot.
 * 
 * Copyright (C) 2006 Geoffrey D. Bennett <g@netcraft.com.au>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <avr/interrupt.h>

#include "config.h"

static uint8_t int0_rise, int1_rise;
static uint8_t int0_duration, int1_duration;

SIGNAL(SIG_INTERRUPT0) {
  uint8_t timer = TCNT2;
  uint8_t mcucr = MCUCR;

  /* rising edge */
  if (mcucr & (1 << ISC00)) {
    /* start looking for falling edge */
    MCUCR = mcucr & ~(1 << ISC00);
    int0_rise = timer;

  /* falling edge */
  } else {
    /* start looking for falling edge */
    MCUCR = mcucr | (1 << ISC00);
    int0_duration = timer - int0_rise;
  }
}

SIGNAL(SIG_INTERRUPT1) {
  uint8_t timer = TCNT2;
  uint8_t mcucr = MCUCR;

  /* rising edge */
  if (mcucr & (1 << ISC10)) {
    /* start looking for falling edge */
    MCUCR = mcucr & ~(1 << ISC10);
    int1_rise = timer;

  /* falling edge */
  } else {
    /* start looking for falling edge */
    MCUCR = mcucr | (1 << ISC10);
    int1_duration = timer - int1_rise;
  }
}

#define VALID_CYCLES 10
#define INVALID_CYCLES 3

/* throttle range 3C 5E 7F */

void get_remote(float *throttle, float *direction) {
  static int valid_cycles = 0;
  static int invalid_cycles = 0;
  static float last_throttle = 0;
  static float last_direction = 0;

  float throttle_value, direction_value;
  uint8_t data_available, data_valid;

  throttle_value = int0_duration;
  int0_duration = 0;
  direction_value = int1_duration;
  int1_duration = 0;

  /* did we see a pulse since last time this was called? */
  data_available = (throttle_value != 0 && direction_value != 0);

  /* convert to -1..1 */
  throttle_value = (throttle_value - config.s.rc_throttle_midpoint)
                   / config.s.rc_throttle_range;
  direction_value = (direction_value - config.s.rc_steer_midpoint)
                   / config.s.rc_steer_range;

  /* is the pulse the correct width? */
  data_valid = (throttle_value >= -1) && (throttle_value <= 1) &&
               (direction_value >= -1) && (direction_value <= 1);

  /* bad pulse width; stop immediately */
  if (data_available && !data_valid) {
    valid_cycles = 0;
    invalid_cycles = 0;
    last_throttle = 0;
    last_direction = 0;
    *throttle = 0;
    *direction = 0;
    return;
  }

  /* currently receiving data? */
  if (valid_cycles >= VALID_CYCLES) {

    /* got a pulse */
    if (data_available) {
      invalid_cycles = 0;

      *throttle = last_throttle = throttle_value;
      *direction = last_direction = direction_value;
      return;

    /* didn't get a pulse */
    } else {

      /* missed too many pulses */
      if (invalid_cycles++ >= INVALID_CYCLES) {
        valid_cycles = 0;
        invalid_cycles = 0;
        last_throttle = 0;
        last_direction = 0;
        *throttle = 0;
        *direction = 0;
        return;
      }

      /* return previous data */
      *throttle = last_throttle;
      *direction = last_direction;
      return;
    }

  /* waiting for enough pulses */
  } else {

    /* got a pulse */
    if (data_available) {
      valid_cycles++;
      invalid_cycles = 0;

    /* didn't get a pulse */
    } else {

      /* count up invalid cycles; if too many, reset valid_cycles */
      if (invalid_cycles < INVALID_CYCLES) {
        invalid_cycles++;
        if (invalid_cycles >= INVALID_CYCLES) {
          valid_cycles = 0;
        }
      }
    }

    *throttle = last_throttle = 0;
    *direction = last_direction = 0;
  }
}
