/*
 * Meta, Geoffrey's self-balancing ride-on robot.
 * 
 * Copyright (C) 2006 Trevor Blackwell <tlb@tlb.org>
 * Copyright (C) 2006-2007 Geoffrey D. Bennett <g@netcraft.com.au>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/interrupt.h>

#include "uart.h"
#include "motor.h"
#include "hardware.h"
#include "config.h"

#define LOG_PITCH_RATE (1 << 0)
#define LOG_AY         (1 << 1)
#define LOG_ANGLE      (1 << 2)
#define LOG_TURN_IN    (1 << 3)
#define LOG_CMD        (1 << 4)
#define LOG_BATT       (1 << 5)
#define LOG_OVERSPEED  (1 << 6)

/* bitmap (set) of variables to log */
uint16_t log_var_set = 0;

/* timeout until we stop logging */
uint8_t log_var_timeout = 0;

float angle;
float rate_bias;
float rate;
float angle_noise;
uint8_t rate_inited;
uint8_t angle_inited;

void extend_count(uint8_t cnt8, uint16_t *cnt16)
{
  uint8_t cnt16l = (uint8_t) *cnt16;
  uint8_t delta = (uint8_t) (cnt8 - cnt16l);
  *cnt16 += delta;
}

float fmax(float a, float b)
{
  if (a >= b) return a;
  return b;
}

float fmin(float a, float b) {
  if (a <= b) return a;
  return b;
}

float flim(float x, float lo, float hi) {
  if (x > hi) return hi;
  if (x < lo) return lo;
  return x;
}

void lpf_update(float *state, float tc, float interval, float input) {
  float frac = interval / tc;
  if (frac > 1.0) {
    frac = 1.0;
  }

  *state = input * frac + *state * (1.0 - frac);
}

void gyro_sample_rate(float in_rate, float interval) {
  if (!rate_inited) {
    rate_bias = in_rate;
    rate_inited = 1;
  }

  rate = in_rate - rate_bias;
  angle += rate * interval;

  rate_bias += rate * 0.3 * interval;
}

void gyro_sample_angle(float in_angle, float interval) {
  float angle_err;

  if (!angle_inited) {
    angle = in_angle;
    angle_inited=1;
  }

  angle_err = in_angle - angle;

  angle += angle_err * interval / (2.0 + angle_noise);
}

static uint16_t tcnt0_ext;
static uint16_t last_balance_tcnt0_ext;

static uint8_t tipped = 1;
static float lpf_angle = 0;
static float lpf_angrate = 0;
static float softstart = 0;
static float balance_torque = 0;
static float cur_speed_est = 0;
static float lpf_turn = 0;
static float overspeed = 0;
static float overspeed_integral = 0;
static float lpf_rc_throttle = 0;

float timer0_seconds_conv;

void timer_init(void) {
  TCCR0 = TCCR0_INIT;
  TCCR2 = TCCR2_INIT;
  timer0_seconds_conv = 1.0/((float)CPUCLK/1024);
}

struct {
  float   turn;
  float   pitch_rate;
  float   reg_5v;
  float   ay;
  float   batt_l;
  float   batt_r;
  uint8_t kill;
  float   rc_throttle;
} in;

/* read all the ADC inputs and do some conversion */
void sample_inputs(void) {
  uint16_t adc0, adc1, adc2, adc3, adc4, adc5;

  adc0 = adc_read(0); /* turn */
  adc1 = adc_read(1); /* gyro */
  adc2 = adc_read(2); /* 2.5V */
  adc3 = adc_read(3); /* accelerometer */
  adc4 = adc_read(4); /* batt left */
  adc5 = adc_read(5); /* batt right */
/*  in.kill = PIND & 0x04; */
  in.kill = 0;

  /* 427 and 568 correspond to 6V and 4.5V respectively*/
  if (adc2 < 427 || adc2 > 568) {
    in.reg_5v = 5.0;
  } else {
    in.reg_5v = 2.5 * 1024 / adc2;
  }

  /* ADXRS401 is not ratiometric -- adjust for ratiometric ADC */
  in.pitch_rate = adc1 * in.reg_5v / 5.0;

  /* should be 2.5V (512) at 0deg/sec */
  in.pitch_rate = (in.pitch_rate - 512.0) * config.s.pitch_rate_conv;

  /* ADXL203 is 2.5V (512) at 0g */
  in.ay = (adc3 - 512.0) * config.s.ay_conv;

  /* 2.5V is neutral */
  in.turn = (adc0 - 512.0) / config.s.turn_range;
  if ((in.turn < -1) || (in.turn > 1)) {
    in.turn = 0;
  }
  in.turn *= config.s.turn_gain;

  /* adjust for ratiometric ADC */
  in.batt_l = adc4 * config.s.batt_conv * in.reg_5v / 5.0;
  in.batt_r = adc5 * config.s.batt_conv * in.reg_5v / 5.0;

  /* sanity check */
  if (in.batt_l < 33.0 || in.batt_l > 42.0) {
    in.batt_l = 36.0;
  }
  if (in.batt_r < 33.0 || in.batt_r > 42.0) {
    in.batt_r = 36.0;
  }
}

/* read the RC inputs */
void sample_rc(void) {
  float throttle, direction;

  get_remote(&throttle, &direction);
  in.rc_throttle = throttle * config.s.rc_throttle_angle_conv;

  if (direction != 0) {
    in.turn = direction * config.s.rc_steer_conv;
  }
}

void balance(void) {
  float cmd, left, right, interval;
  float turn, overspeed_angle, lpf_angle_corrected;

  {
    uint16_t ticks = tcnt0_ext - last_balance_tcnt0_ext;
    last_balance_tcnt0_ext = tcnt0_ext;
    interval = timer0_seconds_conv * (float)ticks;
  }

  sample_inputs();
  sample_rc();

  turn = in.turn / (fabs(cur_speed_est) * config.s.turn_backoff + 1);
  lpf_update(&lpf_turn, config.s.turn_filter, interval, turn);

  if (!tipped) {
    in.ay -= config.s.ay_bal_offset * balance_torque;
  }

  gyro_sample_rate(in.pitch_rate, interval);
  gyro_sample_angle(in.ay, interval);
  lpf_update(&lpf_angle, 0.05, interval, angle);
  lpf_update(&lpf_angrate, 0.04, interval, rate);

  if (!tipped &&
      (lpf_angle > config.s.tip_angle ||
       lpf_angle < -config.s.tip_angle ||
       in.kill)) {
    tipped = 1;
    overspeed_integral = 0;
    softstart = 0.0;
  }

  overspeed = 0.0;

  /* tilt back if we're overspeed */
  if (cur_speed_est > config.s.speed_limit_fwd) {
    overspeed = cur_speed_est - config.s.speed_limit_fwd; /* positive */
    overspeed_integral += config.s.overspeed_i_gain
                          * interval
                          * fmax(0.0, overspeed - 0.08);
    overspeed *= config.s.overspeed_p_gain;
  } else if (cur_speed_est < -config.s.speed_limit_rev) {
    overspeed = cur_speed_est + config.s.speed_limit_rev; /* negative */
    overspeed_integral += config.s.overspeed_i_gain
                          * interval
                          * fmin(0.0, overspeed + 0.08);
    overspeed *= config.s.overspeed_p_gain;

  /* stop tilting back if we're not overspeed*/
  } else if (overspeed_integral > 0.0) {
    /* negative */
    float not_overspeed = fmin(0.0, cur_speed_est - config.s.speed_limit_fwd);
    overspeed_integral += config.s.not_overspeed_i_gain
                          * interval
                          * fmin(-0.02, not_overspeed);
    if (overspeed_integral < 0.0) {
      overspeed_integral = 0.0;
    }
  } else if (overspeed_integral < 0.0) {
    /* positive */
    float not_overspeed = fmin(0.0, cur_speed_est + config.s.speed_limit_rev);
    overspeed_integral += config.s.not_overspeed_i_gain
                          * interval
                          * fmax(0.02, not_overspeed);
    if (overspeed_integral > 0.0) {
      overspeed_integral = 0.0;
    }
  }

  overspeed_integral = flim(overspeed_integral,
                            -config.s.overspeed_angle_limit,
                            config.s.overspeed_angle_limit);

  overspeed_angle = flim(overspeed + overspeed_integral,
                         -config.s.overspeed_angle_limit,
                         config.s.overspeed_angle_limit);

  lpf_update(&lpf_rc_throttle, config.s.rc_throttle_filter,
             interval, in.rc_throttle);

  lpf_angle_corrected = lpf_angle + overspeed_angle + lpf_rc_throttle;

  balance_torque =
    (config.s.p_gain * lpf_angle_corrected +
     config.s.d_gain * lpf_angrate) * softstart;

  angle_noise = fmin(angle_noise, fabs(balance_torque) * 30.0);
  lpf_update(&angle_noise, 0.5, interval, 0.0);

  cmd = (cur_speed_est +
         (1.0 + 0.2 * fabs(cur_speed_est)) * balance_torque);

  if (tipped) {
    cur_speed_est = 0;
  } else {
    cur_speed_est = flim(cur_speed_est +
                           config.s.i_gain * interval * balance_torque,
                         -config.s.hard_speed_lim, config.s.hard_speed_lim);
  }

  cmd = flim(cmd, -config.s.hard_speed_lim, config.s.hard_speed_lim);

  if (!tipped) {
    softstart = fmin(1.0, softstart + config.s.softstart_delay * interval);
  }

  if (tipped && cmd > -0.1 && cmd < 0.1 && !in.kill) {
    tipped = 0;
  }

  left = right = cmd;
  left += lpf_turn;
  right -= lpf_turn;

  left = left * 36.0 / in.batt_l;
  right = right * 36.0 / in.batt_r;

  if (tipped) {
    set_motors(0.0, 0.0);
  } else {
    set_motors(left * config.s.motor_l_gain, right * config.s.motor_r_gain);
  }

  putc('>');
  log_uint8(TCNT0);
  log_uint16(log_var_set);
  if (log_var_set & LOG_PITCH_RATE) { log_float(in.pitch_rate); }
  if (log_var_set & LOG_AY)         { log_float(in.ay); }
  if (log_var_set & LOG_ANGLE)      { log_float(lpf_angle); }
  if (log_var_set & LOG_TURN_IN)    { log_float(in.turn); }
  if (log_var_set & LOG_CMD) {
    log_float(cmd);
    log_float(lpf_turn);
  }
  if (log_var_set & LOG_BATT) {
    log_float(in.reg_5v);
    log_float(in.batt_l);
    log_float(in.batt_r);
  }
  if (log_var_set & LOG_OVERSPEED) {
    log_float(overspeed);
    log_float(overspeed_integral);
    log_float(overspeed_angle);
  }
  putc('<');
  putc('\r');
  putc('\n');
}

void listen(void) {

  /* accumulator */
  static union {
    float    f;
    uint32_t i32;
    uint16_t i16;
    uint8_t  i8;
  } acc;

  static int selected_variable = 0xFF;
  static float new_value;
  static uint8_t magic;

  char c;

  while (c = getc()) {

    /* update accumulator; note that floats need to be input in reverse */
    if (c >= '0' && c <= '9') {
      acc.i32 = (acc.i32 << 4) + (c - '0');
      continue;
    }
    if (c >= 'A' && c <= 'F') {
      acc.i32 = (acc.i32 << 4) + (c - 'A' + 10);
      continue;
    }

    switch (c) {

      /* display accumulator as float, short, or char */
      case 'a':
        log_str("+a");
        log_float(acc.f);
        log_str("\r\n");
        break;
      case 'b':
        log_str("+b");
        log_uint16(acc.i16);
        log_str("\r\n");
        break;
      case 'c':
        log_str("+c");
        log_uint8(acc.i8);
        log_str("\r\n");
        break;

      /* display config variable selected by the accumulator */
      case 'd':
        log_str("+d");
        log_uint8(acc.i8);
        putc('=');
        log_float(config.a[acc.i8]);
        log_str("\r\n");
        break;

      /* save the accumulator as new_value */
      case 'e':
        new_value = acc.f;
        break;

      /* start request to save new_value in the config variable
       * selected by the accumulator */
      case 'f':
        selected_variable = acc.i8;
        if (selected_variable < 0 || selected_variable >= NUM_CONFIG_VAR) {
          log_str("+f!\r\n");
          break;
        }

        magic = TCNT0;

        log_str("+f");
        log_uint8(selected_variable);
        putc('=');
        log_float(new_value);
        putc(' ');
        log_uint8(magic);
        log_str("\r\n");

        break;

      /* finish save request (accumulator needs magic var from prev step) */
      case 'g':
        if (acc.i8 != magic) {
          log_str("+g");
          log_uint8(selected_variable);
          log_str("!\r\n");
          break;
        }

        config.a[selected_variable] = new_value;
        log_str("+g");
        log_uint8(selected_variable);
        putc('=');
        log_float(config.a[selected_variable]);
        log_str("\r\n");
        break;

      /* config and EEPROM requests */
      case 'h':
        /* copy factory defaults to running config */
        if (acc.i8 == 0xA1) {
          config_load_defaults();
          log_str("+h1\r\n");

        /* copy running config to EEPROM */
        } else if (acc.i8 == 0xA2) {
          eeprom_save_config();
          log_str("+h2\r\n");

        /* copy EEPROM to running config */
        } else if (acc.i8 == 0xA3) {
          if (eeprom_load_config()) {
            log_str("+h31\r\n");
          } else {
            log_str("+h30\r\n");
          }
        } else {
          log_str("+h0\r\n");
        }
        break;

      /* set log_var_set from acc.i16 */
      case 'i':
        log_var_set = acc.i16;
        log_var_timeout = 255;
        log_str("+i\r\n");
        break;
    }
  }
}

int main(void) {
  // Disable the AVR's internal watchdog timer.
  wdt_disable();

  /* see hardware */
  DDRA = DDRA_INIT;
  PORTA = PORTA_INIT;

  DDRB = DDRB_INIT;
  PORTB = PORTB_INIT;

  DDRC = DDRC_INIT;
  PORTC = PORTC_INIT;

  DDRD = DDRD_INIT;
  PORTD = PORTD_INIT;

  /* interrupt setup */
  MCUCR = MCUCR_INIT;
  GICR = GICR_INIT;

  init_motor();
  adc_init();
  uart_init();
  timer_init();
  sei();

  log_str("\r\nHello, world!\r\n\r\n");
  uart_flush();

  eeprom_load_config();

  log_str("+GO\r\n");

  last_balance_tcnt0_ext = tcnt0_ext = TCNT0;

  while (1) {
    extend_count(TCNT0, &tcnt0_ext);
    balance();
    listen();

    if (log_var_timeout > 0) {
      if (--log_var_timeout == 0) {
        log_var_set = 0;
      }
    }
  }
}
