/*
 * Meta, Geoffrey's self-balancing ride-on robot.
 * 
 * Copyright (C) 2006 Trevor Blackwell <tlb@tlb.org>
 * Copyright (C) 2006-2007 Geoffrey D. Bennett <g@netcraft.com.au>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <avr/interrupt.h>
#include <avr/io.h>

#include "hardware.h"
#include "uart.h"

#define TXQ_SIZE		128
uint8_t				txq_head;
volatile uint8_t		txq_tail;
static uint8_t			txq_data[TXQ_SIZE];

#define RXQ_SIZE		32
uint8_t				rxq_head;
volatile uint8_t		rxq_tail;
static uint8_t			rxq_data[RXQ_SIZE];

void uart_init(void) {

  /* set baud rate */
  UBRRH = (unsigned char)(UBRR_INIT >> 8);
  UBRRL = (unsigned char)UBRR_INIT;

  /* enable receiver and transmitter; defaults to 8N1, enable rx interrupts */
  UCSRB = (1 << RXEN) | (1 << TXEN) | (1 << RXCIE);

  txq_head = txq_tail = 0;
  rxq_head = rxq_tail = 0;
}

/* USART Data Register Empty Interrupt */
SIGNAL(SIG_UART_DATA) {

  /* if there's nothing left in the txq buffer to transmit,
     disable this interrupt */
  if (txq_head == txq_tail) {
    UCSRB &= ~(1 << UDRIE);
    return;
  }

  /* send one character */
  UDR = txq_data[txq_tail];
  txq_tail = (txq_tail + 1) % TXQ_SIZE;
}

SIGNAL(SIG_UART_RECV) {
  /* receive one character */
  uint8_t c;

  rxq_data[rxq_tail] = UDR;
  rxq_tail = (rxq_tail + 1) % RXQ_SIZE;
}

void putc(uint8_t c) {
  uint8_t next_head = (txq_head + 1) % TXQ_SIZE;

  /* busy loop, waiting for space in buffer */
  while (next_head == txq_tail) { }

  /* put the character into the queue */
  txq_data[txq_head] = c;
  txq_head = next_head;

  /* enable the USART Data Register Empty Interrupt */
  UCSRB |= (1 << UDRIE);
}

uint8_t uart_send_empty(void) {
  return txq_head == txq_tail;
}

uint8_t uart_tx_qlen(void) {
  return (txq_head + TXQ_SIZE - txq_tail) % TXQ_SIZE;
}

void uart_flush(void) {
  while (!uart_send_empty()) {
  }
}

void log_str(const char *s) {
  while (*s) {
    putc(*s++);
  }
}

void log_nybble(uint8_t x) {
  putc(x < 10 ? x + '0' : x - 10 + 'A');
}

void log_uint8(uint8_t x) {
  log_nybble(x >> 4);
  log_nybble(x & 0x0F);
}

void log_uint16(uint16_t x) {
  log_uint8(x >> 8);
  log_uint8(x & 0xFF);
}

void log_float(float x) {
  // Turns out to be same representation as float on i386.
  union {
    float f;
    uint8_t c[4];
  } tmp;
  tmp.f = x;
  log_uint8(tmp.c[0]);
  log_uint8(tmp.c[1]);
  log_uint8(tmp.c[2]);
  log_uint8(tmp.c[3]);
}

char getc(void) {
  char c;

  if (rxq_head == rxq_tail) {
    return 0;
  }

  c = rxq_data[rxq_head];
  rxq_head = (rxq_head + 1) % RXQ_SIZE;
  return c;
}
